// Logging with colour
// ======================================================================================

#define KNRM  "\x1B[0m"
#define KRED  "\x1B[31m"
#define KGRN  "\x1B[32m"
#define KYEL  "\x1B[33m"
#define KBLU  "\x1B[34m"
#define KMAG  "\x1B[35m"
#define KCYN  "\x1B[36m"
#define KWHT  "\x1B[37m"

#define LOG_TYPE_DEV 0
#define LOG_TYPE_INFO 1
#define LOG_TYPE_ALLOC 2
#define LOG_TYPE_FREE 3
#define LOG_TYPE_SUCCESS 4
#define LOG_TYPE_ERROR 5
#define LOG_TYPE_WARN 6

void(*CSR_LOG_REDIR_STDOUT)(int, const char*) = NULL;
int CSR_LOG_ENABLE_STDOUT = 1;

void csr_log_out( FILE *f, int type, const char* prefix, const char* fmt, ... )
{
	if( CSR_LOG_ENABLE_STDOUT )
	{
	   fprintf( f, "%s", prefix );
		va_list args;
		va_start( args, fmt );
		vfprintf( f, fmt, args );
		va_end( args );
		fprintf( f, KNRM ); // reset
	}
	
	// Send to a seperate logging function
	if( CSR_LOG_REDIR_STDOUT )
	{
		char buffer[ 512 ];
		va_list args;
		va_start( args, fmt );
		vsnprintf( buffer, 512, fmt, args );
		va_end( args );
		CSR_LOG_REDIR_STDOUT( type, buffer );
	}
}

#define log_dev(FMT, ...)     csr_log_out( stdout, LOG_TYPE_DEV,     (KNRM "    dev" KWHT "| " KNRM), FMT, ##__VA_ARGS__ )
#define log_info(FMT, ...)    csr_log_out( stdout, LOG_TYPE_INFO,    (KNRM "   info" KWHT "| " KNRM), FMT, ##__VA_ARGS__ )
#define log_alloc(FMT, ...)   csr_log_out( stdout, LOG_TYPE_ALLOC,   (KCYN "  alloc" KWHT "| " KNRM), FMT, ##__VA_ARGS__ )
#define log_free(FMT, ...)    csr_log_out( stdout, LOG_TYPE_FREE,    (KMAG "   free" KWHT "| " KNRM), FMT, ##__VA_ARGS__ )
#define log_success(FMT, ...) csr_log_out( stdout, LOG_TYPE_SUCCESS, (KGRN "success" KWHT "| " KGRN), FMT, ##__VA_ARGS__ )
#define log_error(FMT, ...)   csr_log_out( stderr, LOG_TYPE_ERROR,   (KRED "  error" KWHT "| " KRED), FMT, ##__VA_ARGS__ )
#define log_warn(FMT, ...)    csr_log_out( stdout, LOG_TYPE_WARN,    (KYEL "   warn" KWHT "| " KYEL), FMT, ##__VA_ARGS__ )

#define log_init log_alloc
#define log_dealloc log_free

const char *csr_prog_msg;
void csr_prog_begin( const char *msg )
{
	csr_prog_msg = msg;
}

void csr_prog_update( float const percent )
{
	int const k_track_length = 40;

	printf( "\r %s %d%% |", csr_prog_msg, (int)(percent*100.f) );
	
	// Calculate how many steps travelled
	int cur_pos = ((float)k_track_length * (1.f-percent));
	
	for( int i = 0; i < cur_pos; i ++ )
		printf( " " );
		
	printf( "🚂🚃🚃🚃🚃" );
	
	for( int i = 0; i < k_track_length-cur_pos; i ++ )
		printf( " " );
		
	printf( "|" );
	
	fflush( stdout );
}

void csr_prog_end(void)
{
	csr_prog_update( 1.f );
	printf( "\n" );
}
