// Util
// ==================================================================================================================

#define CSR_PIf 3.14159265358979323846264338327950288f

// Simple min/max replacements
float csr_minf( float a, float b )
{
	return a < b? a: b;
}

float csr_maxf( float a, float b )
{
	return a > b? a: b;
}

int csr_min( int a, int b )
{
	return a < b? a: b;
}

int csr_max( int a, int b )
{
	return a > b? a: b;
}

// Convert double precision vec3 into single
void v3d_v3f( double a[3], float b[3] )
{
	b[0] = a[0];
	b[1] = a[1];
	b[2] = a[2];
}

// Convert degrees to radians
float csr_rad( float deg )
{
	return deg * CSR_PIf / 180.0f;
}

// Vector 2
// ==================================================================================================================

void v2_copy( v2f a, v2f b )
{
	b[0] = a[0]; b[1] = a[1];
}

void v2_minv( v2f a, v2f b, v2f dest ) 
{
	dest[0] = csr_minf(a[0], b[0]);
	dest[1] = csr_minf(a[1], b[1]);
}

void v2_maxv( v2f a, v2f b, v2f dest ) 
{
	dest[0] = csr_maxf(a[0], b[0]);
	dest[1] = csr_maxf(a[1], b[1]);
}

void v2_sub( v2f a, v2f b, v2f d )
{
	d[0] = a[0]-b[0]; d[1] = a[1]-b[1];
}

float v2_cross( v2f a, v2f b )
{
	return a[0] * b[1] - a[1] * b[0];
}

void v2_add( v2f a, v2f b, v2f d )
{
	d[0] = a[0]+b[0]; d[1] = a[1]+b[1];
}

void v2_muls( v2f a, float s, v2f d )
{
	d[0] = a[0]*s; d[1] = a[1]*s;
}

void v2_mul( v2f a, v2f b, v2f d )
{
	d[0] = a[0]*b[0]; d[1] = a[1]*b[1];
}

// Vector 3
// ==================================================================================================================

void v3_zero( v3f a )
{
	a[0] = 0.f; a[1] = 0.f; a[2] = 0.f;
}

void v3_copy( v3f a, v3f b )
{
	b[0] = a[0]; b[1] = a[1]; b[2] = a[2];
}

void v3_add( v3f a, v3f b, v3f d )
{
	d[0] = a[0]+b[0]; d[1] = a[1]+b[1]; d[2] = a[2]+b[2];
}

void v3_sub( v3f a, v3f b, v3f d )
{
	d[0] = a[0]-b[0]; d[1] = a[1]-b[1]; d[2] = a[2]-b[2];
}

void v3_mul( v3f a, v3f b, v3f d )
{
	d[0] = a[0]*b[0]; d[1] = a[1]*b[1]; d[2] = a[2]*b[2];
}

void v3_div( v3f a, v3f b, v3f d )
{
	d[0] = a[0]/b[0]; d[1] = a[1]/b[1];	d[2] = a[2]/b[2];
}

void v3_muls( v3f a, float s, v3f d )
{
	d[0] = a[0]*s; d[1] = a[1]*s; d[2] = a[2]*s;
}

void v3_divs( v3f a, float s, v3f d )
{
	d[0] = a[0]/s; d[1] = a[1]/s; d[2] = a[2]/s;
}

void v3_muladds( v3f a, v3f b, float s, v3f d )
{
	d[0] = a[0]+b[0]*s; d[1] = a[1]+b[1]*s; d[2] = a[2]+b[2]*s;
}

float v3_dot( v3f a, v3f b )
{
	return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

void v3_cross( v3f a, v3f b, v3f d )
{
  d[0] = a[1] * b[2] - a[2] * b[1];
  d[1] = a[2] * b[0] - a[0] * b[2];
  d[2] = a[0] * b[1] - a[1] * b[0];
}

float v3_length2( v3f a )
{
	return v3_dot( a, a );
}

float v3_length( v3f a )
{
	return sqrtf( v3_length2( a ) );
}

float v3_dist2( v3f a, v3f b )
{
	v3f delta;
	v3_sub( a, b, delta );
	return v3_length2( delta );
}

float v3_dist( v3f a, v3f b )
{
	return sqrtf( v3_dist2( a, b ) );
}

void v3_normalize( v3f a )
{
	v3_muls( a, 1.f / v3_length( a ), a );
}

float csr_lerpf( float a, float b, float t )
{
	return a + t*(b-a);
}

void v3_lerp( v3f a, v3f b, float t, v3f d )
{
	d[0] = a[0] + t*(b[0]-a[0]);
	d[1] = a[1] + t*(b[1]-a[1]);
	d[2] = a[2] + t*(b[2]-a[2]);
}

void v3_minv( v3f a, v3f b, v3f dest ) 
{
	dest[0] = csr_minf(a[0], b[0]);
	dest[1] = csr_minf(a[1], b[1]);
	dest[2] = csr_minf(a[2], b[2]);
}

void v3_maxv( v3f a, v3f b, v3f dest ) 
{
	dest[0] = csr_maxf(a[0], b[0]);
	dest[1] = csr_maxf(a[1], b[1]);
	dest[2] = csr_maxf(a[2], b[2]);
}

float v3_minf( v3f a )
{
	return csr_minf( csr_minf( a[0], a[1] ), a[2] );
}

float v3_maxf( v3f a )
{
	return csr_maxf( csr_maxf( a[0], a[1] ), a[2] );
}

void v3_fill( v3f a, float v )
{
	a[0] = v;
	a[1] = v;
	a[2] = v;
}

// Vector 4
// ==================================================================================================================

void v4_copy( v4f a, v4f b )
{
	b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; b[3] = a[3];
}

void v4_zero( v4f a )
{
	a[0] = 0.f; a[1] = 0.f; a[2] = 0.f; a[3] = 0.f;
}

// Matrix 3x3
//======================================================================================================

void m3x3_inv_transpose( m3x3f src, m3x3f dest )
{
	float a = src[0][0], b = src[0][1], c = src[0][2],
			d = src[1][0], e = src[1][1], f = src[1][2],
			g = src[2][0], h = src[2][1], i = src[2][2];

	float det =    1.f / 
					(+a*(e*i-h*f)
                -b*(d*i-f*g)
                +c*(d*h-e*g));
                
	dest[0][0] =  (e*i-h*f)*det;
	dest[1][0] = -(b*i-c*h)*det;
	dest[2][0] =  (b*f-c*e)*det;
	dest[0][1] = -(d*i-f*g)*det;
	dest[1][1] =  (a*i-c*g)*det;
	dest[2][1] = -(a*f-d*c)*det;
	dest[0][2] =  (d*h-g*e)*det;
	dest[1][2] = -(a*h-g*b)*det;
	dest[2][2] =  (a*e-d*b)*det;
}

void m3x3_mulv( m3x3f m, v3f v, v3f d )
{
	v3f res;
	
	res[0] = m[0][0]*v[0] + m[1][0]*v[1] + m[2][0]*v[2];
	res[1] = m[0][1]*v[0] + m[1][1]*v[1] + m[2][1]*v[2];
	res[2] = m[0][2]*v[0] + m[1][2]*v[1] + m[2][2]*v[2];
	
	v3_copy( res, d );
}

// Matrix 4x3
// ==================================================================================================================

#define M4X3_IDENTITY	{{1.0f, 0.0f, 0.0f, },\
								{ 0.0f, 1.0f, 0.0f, },\
								{ 0.0f, 0.0f, 1.0f, },\
								{ 0.0f, 0.0f, 0.0f }}

void m4x3_to_3x3( m4x3f a, m3x3f b )
{
	v3_copy( a[0], b[0] );
	v3_copy( a[1], b[1] );
	v3_copy( a[2], b[2] );
}

void m4x3_copy( m4x3f a, m4x3f b )
{
	v3_copy( a[0], b[0] );
	v3_copy( a[1], b[1] );
	v3_copy( a[2], b[2] );
	v3_copy( a[3], b[3] );
}

void m4x3_identity( m4x3f a )
{
	m4x3f id = M4X3_IDENTITY;
	m4x3_copy( id, a );
}

void m4x3_mul( m4x3f a, m4x3f b, m4x3f d ) 
{
	float 
	a00 = a[0][0], a01 = a[0][1], a02 = a[0][2],
	a10 = a[1][0], a11 = a[1][1], a12 = a[1][2],
	a20 = a[2][0], a21 = a[2][1], a22 = a[2][2],
	a30 = a[3][0], a31 = a[3][1], a32 = a[3][2],
	b00 = b[0][0], b01 = b[0][1], b02 = b[0][2],
	b10 = b[1][0], b11 = b[1][1], b12 = b[1][2],
	b20 = b[2][0], b21 = b[2][1], b22 = b[2][2],
	b30 = b[3][0], b31 = b[3][1], b32 = b[3][2];
	
	d[0][0] = a00*b00 + a10*b01 + a20*b02;
	d[0][1] = a01*b00 + a11*b01 + a21*b02;
	d[0][2] = a02*b00 + a12*b01 + a22*b02;
	d[1][0] = a00*b10 + a10*b11 + a20*b12;
	d[1][1] = a01*b10 + a11*b11 + a21*b12;
	d[1][2] = a02*b10 + a12*b11 + a22*b12;
	d[2][0] = a00*b20 + a10*b21 + a20*b22;
	d[2][1] = a01*b20 + a11*b21 + a21*b22;
	d[2][2] = a02*b20 + a12*b21 + a22*b22;
	d[3][0] = a00*b30 + a10*b31 + a20*b32 + a30;
	d[3][1] = a01*b30 + a11*b31 + a21*b32 + a31;
	d[3][2] = a02*b30 + a12*b31 + a22*b32 + a32;
}

void m4x3_mulv( m4x3f m, v3f v, v3f d ) 
{
	v3f res;
  
	res[0] = m[0][0]*v[0] + m[1][0]*v[1] + m[2][0]*v[2] + m[3][0];
	res[1] = m[0][1]*v[0] + m[1][1]*v[1] + m[2][1]*v[2] + m[3][1];
	res[2] = m[0][2]*v[0] + m[1][2]*v[1] + m[2][2]*v[2] + m[3][2];

	v3_copy( res, d );
}

// Affine transforms

void m4x3_translate( m4x3f m, v3f v )
{
	v3_muladds( m[3], m[0], v[0], m[3] );
	v3_muladds( m[3], m[1], v[1], m[3] );
	v3_muladds( m[3], m[2], v[2], m[3] );
}

void m4x3_scale( m4x3f m, float s )
{
	v3_muls( m[0], s, m[0] );
	v3_muls( m[1], s, m[1] );
	v3_muls( m[2], s, m[2] );
}

void m4x3_rotate_x( m4x3f m, float angle )
{
	m4x3f t = M4X3_IDENTITY;
	float c, s;

	c = cosf( angle );
	s = sinf( angle );

	t[1][1] =  c;
	t[1][2] =  s;
	t[2][1] = -s;
	t[2][2] =  c;

	m4x3_mul( m, t, m );
}

void m4x3_rotate_y( m4x3f m, float angle )
{
	m4x3f t = M4X3_IDENTITY;
	float c, s;

	c = cosf( angle );
	s = sinf( angle );

	t[0][0] =  c;
	t[0][2] = -s;
	t[2][0] =  s;
	t[2][2] =  c;

	m4x3_mul( m, t, m );
}

void m4x3_rotate_z( m4x3f m, float angle )
{
	m4x3f t = M4X3_IDENTITY;
	float c, s;

	c = cosf( angle );
	s = sinf( angle );

	t[0][0] =  c;
	t[0][1] =  s;
	t[1][0] = -s;
	t[1][1] =  c;

	m4x3_mul( m, t, m );
}

// Warning: These functions are unoptimized..
void m4x3_expand_aabb_point( m4x3f m, boxf box, v3f point )
{
	v3f v;
	m4x3_mulv( m, point, v );

	v3_minv( box[0], v, box[0] );
	v3_maxv( box[1], v, box[1] );
}

void box_concat( boxf a, boxf b )
{
	v3_minv( a[0], b[0], a[0] );
	v3_maxv( a[1], b[1], a[1] );
}

void box_copy( boxf a, boxf b )
{
	v3_copy( a[0], b[0] );
	v3_copy( a[1], b[1] );
}

void m4x3_transform_aabb( m4x3f m, boxf box )
{
	v3f a; v3f b;
	
	v3_copy( box[0], a );
	v3_copy( box[1], b );
	v3_fill( box[0],  INFINITY );
	v3_fill( box[1], -INFINITY );

	m4x3_expand_aabb_point( m, box, a );
	m4x3_expand_aabb_point( m, box, (v3f){ a[0], b[1], a[2] } );
	m4x3_expand_aabb_point( m, box, (v3f){ b[0], a[1], a[2] } );
	m4x3_expand_aabb_point( m, box, (v3f){ b[0], b[1], a[2] } );
	m4x3_expand_aabb_point( m, box, b );
	m4x3_expand_aabb_point( m, box, (v3f){ a[0], b[1], b[2] } );
	m4x3_expand_aabb_point( m, box, (v3f){ b[0], a[1], b[2] } );
	m4x3_expand_aabb_point( m, box, (v3f){ b[0], b[1], b[2] } );
}

// Planes (double precision)
// ==================================================================================================================

void tri_to_plane( double a[3], double b[3], double c[3], double p[4] )
{
	double edge0[3];
	double edge1[3];
	double l;
	
	edge0[0] = b[0] - a[0];
	edge0[1] = b[1] - a[1];
	edge0[2] = b[2] - a[2];
	
	edge1[0] = c[0] - a[0];
	edge1[1] = c[1] - a[1];
	edge1[2] = c[2] - a[2];
	
	p[0] = edge0[1] * edge1[2] - edge0[2] * edge1[1];
	p[1] = edge0[2] * edge1[0] - edge0[0] * edge1[2];
	p[2] = edge0[0] * edge1[1] - edge0[1] * edge1[0];
	
	l = sqrt(p[0] * p[0] + p[1] * p[1] + p[2] * p[2]);
	p[3] = (p[0] * a[0] + p[1] * a[1] + p[2] * a[2]) / l;
	
	p[0] = p[0] / l;
	p[1] = p[1] / l;
	p[2] = p[2] / l;
}

int plane_intersect( double a[4], double b[4], double c[4], double p[4] )
{
	double const epsilon = 1e-8f;
	
	double x[3];
	double d;
	
	x[0] = a[1] * b[2] - a[2] * b[1];
	x[1] = a[2] * b[0] - a[0] * b[2];
	x[2] = a[0] * b[1] - a[1] * b[0];
	
	d = x[0] * c[0] + x[1] * c[1] + x[2] * c[2];
	
	if( d < epsilon && d > -epsilon ) return 0;
	
	p[0] = (b[1] * c[2] - b[2] * c[1]) * -a[3];
	p[1] = (b[2] * c[0] - b[0] * c[2]) * -a[3];
	p[2] = (b[0] * c[1] - b[1] * c[0]) * -a[3];
	
	p[0] += (c[1] * a[2] - c[2] * a[1]) * -b[3];
	p[1] += (c[2] * a[0] - c[0] * a[2]) * -b[3];
	p[2] += (c[0] * a[1] - c[1] * a[0]) * -b[3];
	
	p[0] += (a[1] * b[2] - a[2] * b[1]) * -c[3];
	p[1] += (a[2] * b[0] - a[0] * b[2]) * -c[3];
	p[2] += (a[0] * b[1] - a[1] * b[0]) * -c[3];
	
	p[0] = -p[0] / d;
	p[1] = -p[1] / d;
	p[2] = -p[2] / d;
	
	return 1;
}

double plane_polarity( double p[4], double a[3] )
{
	return 
	(a[0] * p[0] + a[1] * p[1] + a[2] * p[2])
	-(p[0]*p[3] * p[0] + p[1]*p[3] * p[1] + p[2]*p[3] * p[2])
	;
}
