#ifndef CSR_MEM_H
#define CSR_MEM_H

#include "csrTypes.h"

void *csr_malloc( size_t size )
{
	void *ret = malloc( size );
	if( !ret )
		exit(1);
	return ret;
}

void *csr_calloc( size_t size )
{
	void *ret = calloc( 1, size );
	if( !ret )
		exit(1);
	return ret;
}

void *csr_sb_raw( void *arr )
{
	return ((u32 *)arr)-3;
}

u32 csr_sb_count( void *arr )
{
	return arr? ((u32 *)csr_sb_raw( arr ))[1]: 0;
}

u32 csr_sb_cap( void *arr )
{
	return arr? ((u32 *)csr_sb_raw( arr ))[0]: 0;
}

void *csr_sb_use( void *arr )
{
	u32 *raw = (u32 *)csr_sb_raw( arr );
	void *data = ((u8 *)arr) + raw[1]*raw[2];

	raw[1] ++;
	
	return data;
}

void csr_sb_inc( void *arr, u32 amt )
{
	u32 *raw = (u32 *)csr_sb_raw( arr );
	raw[1] += amt;
}

__attribute__((warn_unused_result))
void *csr_sb_reserve( void *arr, u32 amt, u32 esize )
{
	u32 cap = arr? csr_sb_cap( arr ): 0;
	u32 count = arr? csr_sb_count( arr ): 0;
	
	if( count + amt > cap )
	{
		cap = csr_max( count + amt, cap * 2 );
		u32 *new_ptr = (u32 *)realloc( arr? csr_sb_raw( arr ): NULL, cap * esize + 3*sizeof(u32) );
		
		if( !new_ptr )
			abort();
		
		new_ptr[0] = cap;
		new_ptr[1] = count;
		new_ptr[2] = esize;
		
		return (void *)(new_ptr+3);
	}
	else
	{
		return arr;
	}
}

void csr_sb_free( void *arr )
{
	if( arr )
	{
		free( csr_sb_raw( arr ) );
	}
}

void csr_sb_clear( void *arr )
{
	if( arr )
	{
		((u32 *)csr_sb_raw( arr ))[1] = 0;
	}
}

// djb2 - Dan Bernstein
unsigned long djb2( unsigned char const *str )
{
	unsigned long hash = 5381;
	int c;

	while((c = *str++))
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */

	return hash;
}

#endif
